// Copyright (c) Shenyang Leading Edge Intelligent Technology Co., Ltd. All rights reserved.

package container

import (
	"LEIT_PM/web/models"
	"fmt"
	"github.com/kataras/iris/v12"
	"github.com/kataras/iris/v12/context"
	"github.com/kataras/iris/v12/core/router"
	"reflect"
	"strings"
)

const (
	ServicesPackageName            = "services"
	ServiceNameSuffix              = "Service"
	PathSeparator                  = "/"
	HttpPost            HttpMethod = "POST"
	HttpPut             HttpMethod = "PUT"
	HttpGet             HttpMethod = "GET"
	HttpDelete          HttpMethod = "DELETE"
)

var ServiceNameSuffixLength = len([]rune(ServiceNameSuffix))

type HttpMethod string

type (
	// 调用器
	// 参数
	// 1.参数列表
	// 返回值:
	// 1.返回值列表
	Caller = func([]reflect.Value) []reflect.Value
	// 参数解析器
	// 参数
	// 1.Web上下文
	// 返回值:
	// 1.解析出的参数值
	// 2.错误
	ParametersParser = func(context.Context) ([]interface{}, error)
	// Web解析器,根据方法元数据解析出处理web请求的组件和信息
	// 参数
	// 1.服务名称
	// 2.服务方法类型
	// 返回值:
	// 1.HTTP请求方法
	// 2.参数解析器
	// 3.附加路径
	// 4.错误
	WebParser = func(string, Method) (HttpMethod, ParametersParser, string, error)
)

// *SessionContext类型常量
var sessionContextType = reflect.TypeOf((*SessionContext)(nil))

// 空参数列表
var emptyParameters []reflect.Value

// int 类型常量
var IntType = reflect.TypeOf(0)

// int8 类型常量
var Int8Type = reflect.TypeOf(int8(0))

// int16 类型常量
var Int16Type = reflect.TypeOf(int16(0))

// int32 类型常量
var Int32Type = reflect.TypeOf(int32(0))

// int64 类型常量
var Int64ype = reflect.TypeOf(int64(0))

// uint 类型常量
var UintType = reflect.TypeOf(uint(0))

// uint8 类型常量
var Uint8Type = reflect.TypeOf(uint8(0))

// uint16 类型常量
var Uint16Type = reflect.TypeOf(uint16(0))

// uint32 类型常量
var Uint32Type = reflect.TypeOf(uint32(0))

// uint64 类型常量
var Uint64Type = reflect.TypeOf(uint64(0))

// bool 类型常量
var BoolType = reflect.TypeOf(false)

// string 类型常量
var StringType = reflect.TypeOf("")

// float32 类型常量
var Float32Type = reflect.TypeOf(float32(0))

// float64 类型常量
var Float64Type = reflect.TypeOf(float64(0))

// JSON数组 类型常量
var JsonListType = reflect.TypeOf([]interface{}{})

// JSON对象 类型常量
var JsonObjectType = reflect.TypeOf(map[string]interface{}{})

// JSON对象数组 类型常量
var JsonObjectListType = reflect.TypeOf([]map[string]interface{}{})

// *error 类型常量
var ErrorType = reflect.TypeOf((*error)(nil)).Elem()

// *RequestContext类型常量
var RequestContextType = reflect.TypeOf((*RequestContext)(nil))

// 无会话(XORM会话)上下文
var ContextOfNoSession = NewNoSessionBrokerBuilder()

// 新会话(XORM会话)上下文
var ContextOfNewSession = NewNewSessionBrokerBuilder()

// 新事务代理创建器上下文
var ContextOfNewTransaction = NewNewTransactionBrokerBuilder()

func IsEmptyInterface(theType reflect.Type) bool {
	return theType.Kind() == reflect.Interface && theType.NumMethod() <= 0
}

//
//// 组件信息管理器
//var currentInformationManager = NewInformationManager()

// 解析组件工厂,返回:组件工厂Value,组件模块名,组件接口,错误
// 参数
// 1.工厂方法
// 返回值:
// 1.组件工厂
// 2.组件接口类型
// 异常:
// 1.工厂不能为空
// 2.工厂必须是func
// 3.工厂不能有参数
// 4.工厂必值需有返回值
// 5.工厂只能有一个返回值
// 6.工厂的返回值必需是interface
func parseFactory(factory interface{}) (Factory, Interface) {
	if factory == nil {
		panic(fmt.Sprintf("工厂不能为空!"))
	}
	valueOfFactory := reflect.ValueOf(factory)
	typeOfFactory := valueOfFactory.Type()
	if typeOfFactory.Kind() != reflect.Func {
		panic(fmt.Sprintf("工厂必须是func!"))
	}
	if typeOfFactory.NumIn() != 0 {
		panic(fmt.Sprintf("工厂不能有参数!"))
	}
	if typeOfFactory.NumOut() < 1 {
		panic(fmt.Sprintf("工厂必值需有返回值!"))
	}
	if typeOfFactory.NumOut() > 1 {
		panic(fmt.Sprintf("工厂只能有一个返回值!"))
	}
	if typeOfFactory.Out(0).Kind() != reflect.Interface {
		panic(fmt.Sprintf("工厂的返回值必需是interface!"))
	}
	componentType := typeOfFactory.Out(0)
	return Factory(valueOfFactory), componentType
}

// Caller相等性判断
// 参数
// 1.比较对象1
// 3.比较对象2
// 返回值:
// 1.判断结果
func CallerEqual(caller1 Caller, caller2 Caller) bool {
	if caller1 == nil {
		if caller2 == nil {
			return true
		} else {
			return false
		}
	} else {
		if caller2 == nil {
			return false
		} else {
			return reflect.ValueOf(caller1).Pointer() == reflect.ValueOf(caller2).Pointer()
		}
	}
}

var GlobalInformations = NewInformationManager()

// 注册路由
// 参数:
// 1.路由分组
// 2.默认Web解析器
// 3.组件信息管理器
// 4.事务句柄工厂
// 返回值:
// 1.错误
// 异常:
// 1.组件信息管理器不能为空
// 2.事务句柄工厂不能为空
// 3.服务不在包或他的子包之下
// 4.组件接口名称的后缀不正确
func RegisterRoutes(party router.Party, defaultWebParser WebParser, informationManager InformationManager, transactionHandlerFactory TransactionHandlerFactory) {
	sessionManager := NewSessionManager(informationManager, transactionHandlerFactory)
	groups := make(map[string][]*ServiceInformation, 10)
	for _, information := range informationManager.Items() {
		if serviceInformation, ok := information.(*ServiceInformation); ok {
			packagePath := serviceInformation.Interface().PkgPath()
			group, ok := groups[packagePath]
			if !ok {
				group = make([]*ServiceInformation, 0, 10)
			}
			groups[packagePath] = append(group, serviceInformation)
		}
		for packagePath, group := range groups {
			packageList := strings.Split(packagePath, PathSeparator)
			moudlePath := ""
			for index, packageName := range packageList {
				if strings.ToLower(packageName) == ServicesPackageName {
					moudlePath = strings.Join(packageList[index+1:], PathSeparator)
					break
				}
			}
			if moudlePath == "" {
				panic(fmt.Sprintf("包%s不在包%s或他的子包之下!", packagePath, ServicesPackageName))
			}
			moudleParty := party.Party(fmt.Sprintf("/%s", strings.ToLower(moudlePath)))
			for _, serviceInformation := range group {
				serviceType := serviceInformation.Interface()
				serviceName := serviceType.Name()
				if !strings.HasSuffix(serviceName, ServiceNameSuffix) {
					panic(fmt.Sprintf("组件%s接口名称的后缀不是%s!", serviceName, ServiceNameSuffix))
				}
				runesOfServiceName := []rune(serviceName)
				serviceParty := moudleParty.Party(fmt.Sprintf("/%s", strings.ToLower(string(runesOfServiceName[:len(runesOfServiceName)-ServiceNameSuffixLength]))))
				for _, serviceMethodInformation := range serviceInformation.Methods() {
					methodName := serviceMethodInformation.MethodName()
					webParser := serviceMethodInformation.WebParser()
					if webParser == nil {
						webParser = defaultWebParser
					}

					httpMethod, parametersParser, addedPath, err := webParser(fmt.Sprintf("%s.%s", serviceName, methodName), serviceMethodInformation.MethodType())
					if err != nil {
						panic(err.Error())
					}

					var irisRouter = func(context context.Context) {

						//user, signature, ok := jwts.ParseTokenAndSignature(context)
						user := &models.Usertab{
							Pid:    100,
							Userid: "sa",
							Role:   "",
							RoleId: 0,
						}
						signature := ""
						ok := true
						if !ok {
							RequestError(context, "解析参数失败")
							return
						}
						parameters, err := parametersParser(context)
						if err != nil {
							ContainerError(context, iris.StatusInternalServerError, err)
							return
						}
						requestContext, err := NewRequestContext(context)
						if err != nil {
							ContainerError(context, iris.StatusInternalServerError, err)
							return
						}
						sessionId := user.Userid + signature
						session, err := sessionManager.GetSession(sessionId, user)
						if err != nil {
							ContainerError(context, iris.StatusInternalServerError, err)
							return
						}
						handlers := session.Handlers()
						componentHandler := handlers.Handler(serviceType)
						var serviceHandler *ServiceHandler
						serviceHandler, ok = componentHandler.(*ServiceHandler)
						if !ok {
							RequestError(context, "竟然没找到服务组件句柄!")
							return
						}
						caller := serviceHandler.Method(methodName).Caller()
						if caller == nil {
							RequestError(context, "没找到方法调用器!")
							return
						}
						defer func() {
							if err := recover(); err != nil {
								InternalServerError(context)
								return
							}
						}()
						parameterValues := make([]reflect.Value, len(parameters)+1)
						parameterValues[0] = reflect.ValueOf(requestContext)
						for i, parameter := range parameters {
							parameterValues[i+1] = reflect.ValueOf(parameter)
						}
						returnValues := caller(parameterValues)
						var errValue interface{}
						var result interface{}
						if len(returnValues) > 1 {
							errValue = returnValues[1].Interface()
							result = returnValues[0].Interface()
						} else {
							errValue = returnValues[0].Interface()
							result = nil
						}
						if errValue != nil {
							RemotingError(context, errValue.(error))
						} else {
							Ok(context, result)
						}
					}
					serviceParty.Handle(string(httpMethod), fmt.Sprintf("/%s", strings.ToLower(methodName)+addedPath), irisRouter)
				}
			}
		}
	}
}

func Ok(ctx iris.Context, result interface{}) {
	ctx.StatusCode(iris.StatusOK)
	ctx.JSON(map[string]interface{}{
		"HasError": false,
		"Result":   result,
		"Error":    nil,
	})
}

func RemotingError(ctx iris.Context, err error) {
	ctx.StatusCode(iris.StatusOK)
	ctx.JSON(map[string]interface{}{
		"HasError": true,
		"Result":   nil,
		"Error":    err.Error(),
	})
}

func RequestError(ctx iris.Context, msg string) {
	ctx.StatusCode(iris.StatusInternalServerError)
	ctx.JSON(map[string]interface{}{
		"HasError": true,
		"Result":   nil,
		"Error":    msg,
	})
}

func InternalServerError(ctx iris.Context) {
	ctx.StatusCode(iris.StatusInternalServerError)
	ctx.JSON(map[string]interface{}{
		"HasError": true,
		"Result":   nil,
		"Error":    "内部错误",
	})
}

func ContainerError(ctx iris.Context, status int, err error) {
	ctx.StatusCode(status)
	ctx.JSON(map[string]interface{}{
		"HasError": true,
		"Result":   nil,
		"Error":    err.Error(),
	})
}
