// Copyright (c) Shenyang Leading Edge Intelligent Technology Co., Ltd. All rights reserved.

package container

import (
	"fmt"
	"github.com/go-xorm/xorm"
	"github.com/kataras/iris/v12/context"
	"github.com/kataras/iris/v12/core/memstore"
	"io"
	"mime/multipart"
	"net/http"
	"reflect"
	"testing"
	"time"
)

func FactoryWithoutResult() {
}

type TestElement interface {
}

type TestElementImplement struct {
}

func FactoryOfNullTestElement() TestElement {
	return nil
}

func FactoryOfElementTestInterfaceWithContent(string) TestElement {
	return &TestElementImplement{}
}

func FactoryOfTestElementImplement() *TestElementImplement {
	return &TestElementImplement{}
}

func FactoryOfTestElement() TestElement {
	return &TestElementImplement{}
}

func FactoryWithTooManyResults() (TestElement, TestService) {
	return &TestElementImplement{}, &TestServiceImplement{}
}

var testElementType = reflect.TypeOf((*TestElement)(nil)).Elem()

var elementInstance = Instance(reflect.ValueOf(FactoryOfTestElement()))

type TestService interface {
	String() (string, error)
	StringWithContext(*RequestContext) (string, error)
	StringNotFirstContext(string, *RequestContext) (string, error)
	String2() (string, string, error)
	String3(*RequestContext, *RequestContext) error
	NoError() string
}

type TestToString interface {
	String() string
}

type TestServiceImplement struct {
	content string
}

func (impl *TestServiceImplement) String() (string, error) {
	return impl.content, nil
}

func (impl *TestServiceImplement) StringWithContext(_ *RequestContext) (string, error) {
	return impl.content, nil
}

func (impl *TestServiceImplement) StringNotFirstContext(_ string, _ *RequestContext) (string, error) {
	return "", nil
}

func (impl *TestServiceImplement) String2() (string, string, error) {
	return impl.content, impl.content, nil
}

func (impl *TestServiceImplement) String3(_ *RequestContext, _ *RequestContext) error {
	return nil
}

func (impl *TestServiceImplement) NoError() string {
	return impl.content
}

func FactoryOfTestService() TestService {
	return &TestServiceImplement{}
}

func FactoryOfServiceTestInterfaceWithContent(content string) TestService {
	return &TestServiceImplement{content}
}

type TestServiceImplementWithContext struct {
	context *SessionContext
}

func (impl *TestServiceImplementWithContext) String() (string, error) {
	return "", nil
}

func (impl *TestServiceImplementWithContext) StringWithContext(_ *RequestContext) (string, error) {
	return "", nil
}

func (impl *TestServiceImplementWithContext) StringNotFirstContext(_ string, _ *RequestContext) (string, error) {
	return "", nil
}

func (impl *TestServiceImplementWithContext) String2() (string, string, error) {
	return "", "", nil
}

func (impl *TestServiceImplementWithContext) String3(_ *RequestContext, _ *RequestContext) error {
	return nil
}

func (impl *TestServiceImplementWithContext) NoError() string {
	return ""
}

func (impl *TestServiceImplementWithContext) SetContext(context *SessionContext) {
	impl.context = context
}

type TestWithTooManyInSetContext interface {
	SetContext(*SessionContext, string)
}

type TestWithTooManyInSetContextImplement struct {
	context *SessionContext
}

func (impl *TestWithTooManyInSetContextImplement) SetContext(_ *SessionContext, _ string) {
}

func FactoryOfTestWithTooManyInSetContext() TestWithTooManyInSetContext {
	return &TestWithTooManyInSetContextImplement{}
}

type TestWithReturnValueSetContext interface {
	SetContext(*SessionContext) error
}

type TestWithReturnValueSetContextImplement struct {
	context *SessionContext
}

func (impl *TestWithReturnValueSetContextImplement) SetContext(context *SessionContext) error {
	impl.context = context
	return nil
}

func FactoryOfTestWithReturnValueSetContext() TestWithReturnValueSetContext {
	return &TestWithReturnValueSetContextImplement{}
}

type TestWithErrorInSetContext interface {
	SetContext(string)
}

type TestWithErrorInSetContextImplement struct {
	context *SessionContext
}

func (impl *TestWithErrorInSetContextImplement) SetContext(_ string) {
}

func FactoryOfTestWithErrorInSetContext() TestWithErrorInSetContext {
	return &TestWithErrorInSetContextImplement{}
}

type TestWithNoInSetContext interface {
	SetContext()
}

type TestWithNoInSetContextImplement struct {
	context *SessionContext
}

func (impl *TestWithNoInSetContextImplement) SetContext() {
}

func FactoryOfTestWithNoInSetContext() TestWithNoInSetContext {
	return &TestWithNoInSetContextImplement{}
}

type TestToStringImplement struct {
	context *SessionContext
}

func (impl TestToStringImplement) String() string {
	return ""
}

func FactoryOfTestServiceWithContext() TestService {
	return &TestServiceImplementWithContext{}
}

func FactoryOfTestServiceImplementWithContext() *TestServiceImplementWithContext {
	return &TestServiceImplementWithContext{}
}

func FactoryOfTestServiceImplementWithContextStruct() TestServiceImplementWithContext {
	return TestServiceImplementWithContext{}
}

func FactoryOfNullTestService() TestService {
	return nil
}

func FactoryOfNullTestServiceImplementWithContext() *TestServiceImplementWithContext {
	return nil
}

func FactoryOfInt() *int {
	intValue := 123
	return &intValue
}

func FactoryOfTestToString() TestToString {
	return TestToStringImplement{}
}

var testServiceType = reflect.TypeOf((*TestService)(nil)).Elem()

var serviceInstance = Instance(reflect.ValueOf(FactoryOfTestService()))

type TestCallerBrokerBuilderImplementWithError struct {
	checkingError error
	buildingError error
}

func (impl *TestCallerBrokerBuilderImplementWithError) Check(_ Method) error {

	return impl.checkingError
}

func (impl *TestCallerBrokerBuilderImplementWithError) Build(_ *SessionContext, _ Caller) (Caller, error) {

	return nil, impl.buildingError
}

type TestCallerBrokerBuilderImplement struct {
}

func (impl *TestCallerBrokerBuilderImplement) Check(_ Method) error {

	return nil
}

func (impl *TestCallerBrokerBuilderImplement) Build(_ *SessionContext, caller Caller) (Caller, error) {

	return caller, nil
}

type TransactionHandlerMock struct {
	closeNumberOfTimes  int
	beginNumberOfTimes  int
	beginError          error
	commitNumberOfTimes int
	commitError         error
	session             *xorm.Session
}

func (mock *TransactionHandlerMock) Close() {
	mock.closeNumberOfTimes++
}

func (mock *TransactionHandlerMock) Begin() error {
	mock.beginNumberOfTimes++
	return mock.beginError
}

func (mock *TransactionHandlerMock) Commit() error {
	mock.commitNumberOfTimes++
	return mock.commitError
}

func (mock *TransactionHandlerMock) Session() *xorm.Session {
	return mock.session
}

type TransactionHandlerFactoryMock struct {
	handler     *TransactionHandlerMock
	createError error
}

// @Reference LEIT_PM/container/TransactionHandlerFactory.Create
func (mock *TransactionHandlerFactoryMock) Create() (TransactionHandler, error) {
	return mock.handler, mock.createError
}

var transactionHandlerFactoryMock TransactionHandlerFactory = &TransactionHandlerFactoryMock{}

func TestContainer_ParseFactory(t *testing.T) {

	_, _ = parseFactory(FactoryOfTestElement)
}

func TestContainer_ParseFactory_NullError(t *testing.T) {
	defer func() {
		if err := recover(); err != "工厂不能为空!" {
			t.Fatalf("意外错误:%s", err)
		}
	}()

	_, _ = parseFactory(nil)
	t.Fatalf("意外的没有错误!")
}

func TestContainer_ParseFactory_FactoryTypeError(t *testing.T) {
	defer func() {
		if err := recover(); err != "工厂必须是func!" {
			t.Fatalf("意外错误:%s", err)
		}
	}()

	_, _ = parseFactory(reflect.TypeOf(FactoryOfTestElement()))
	t.Fatalf("意外的没有错误!")
}

func TestContainer_ParseFactory_ParameterCountError(t *testing.T) {
	defer func() {
		if err := recover(); err != "工厂不能有参数!" {
			t.Fatalf("意外错误:%s", err)
		}
	}()

	_, _ = parseFactory(FactoryOfElementTestInterfaceWithContent)
	t.Fatalf("意外的没有错误!")
}

func TestContainer_ParseFactory_NoReturnValueError(t *testing.T) {
	defer func() {
		if err := recover(); err != "工厂必值需有返回值!" {
			t.Fatalf("意外错误:%s", err)
		}
	}()

	_, _ = parseFactory(FactoryWithoutResult)
	t.Fatalf("意外的没有错误!")
}

func TestContainer_ParseFactory_TooManyResultsError(t *testing.T) {
	defer func() {
		if err := recover(); err != "工厂只能有一个返回值!" {
			t.Fatalf("意外错误:%s", err)
		}
	}()

	_, _ = parseFactory(FactoryWithTooManyResults)
	t.Fatalf("意外的没有错误!")
}

func TestContainer_ParseFactory_ReturnValueTypeError(t *testing.T) {
	defer func() {
		if err := recover(); err != "工厂的返回值必需是interface!" {
			t.Fatalf("意外错误:%s", err)
		}
	}()

	_, _ = parseFactory(FactoryOfTestElementImplement)
	t.Fatalf("意外的没有错误!")
}

func call1(_ []reflect.Value) []reflect.Value {
	return []reflect.Value{reflect.ValueOf(1)}
}

func call2(_ []reflect.Value) []reflect.Value {
	return []reflect.Value{reflect.ValueOf(2)}
}

func TestContainer_Caller_Equal_NullCaller1AndCaller2(t *testing.T) {

	if !CallerEqual(nil, nil) {
		t.Fatalf("意外的不相等!")
	}
}

func TestContainer_Caller_Equal_NullCaller1(t *testing.T) {

	if CallerEqual(nil, call1) {
		t.Fatalf("意外的相等!")
	}
}

func TestContainer_Caller_Equal_NullCaller2(t *testing.T) {

	if CallerEqual(call1, nil) {
		t.Fatalf("意外的相等!")
	}
}

func TestContainer_Caller_Equal_NotNull(t *testing.T) {

	var caller1 = call1
	var caller2 = call1
	var caller3 = call2
	if !CallerEqual(caller1, caller2) {
		t.Fatalf("意外的不相等!")
	}
	if CallerEqual(caller2, caller3) {
		t.Fatalf("意外的相等!")
	}
	call3 := func(_ []reflect.Value) []reflect.Value {
		return []reflect.Value{reflect.ValueOf(3)}
	}
	call4 := func(_ []reflect.Value) []reflect.Value {
		return []reflect.Value{reflect.ValueOf(4)}
	}
	var caller4 = call3
	var caller5 = call3
	var caller6 = call4
	if !CallerEqual(caller4, caller5) {
		t.Fatalf("意外的不相等!")
	}
	if CallerEqual(caller5, caller6) {
		t.Fatalf("意外的相等!")
	}
}

type Synchronizer interface {
	Awake()
	Wait()
}

type SynchronizerImplement struct {
	core chan interface{}
}

func NewSynchronizer() Synchronizer {
	return &SynchronizerImplement{make(chan interface{}, 1)}
}

func (impl SynchronizerImplement) Awake() {
	impl.core <- 1
}

func (impl SynchronizerImplement) Wait() {
	<-impl.core
}

type ContextMock struct {
}

func (mock *ContextMock) BeginRequest(http.ResponseWriter, *http.Request)     {}
func (mock *ContextMock) EndRequest()                                         {}
func (mock *ContextMock) ResponseWriter() context.ResponseWriter              { return nil }
func (mock *ContextMock) ResetResponseWriter(_ context.ResponseWriter)        {}
func (mock *ContextMock) Request() *http.Request                              { return nil }
func (mock *ContextMock) ResetRequest(_ *http.Request)                        {}
func (mock *ContextMock) SetCurrentRouteName(_ string)                        {}
func (mock *ContextMock) GetCurrentRoute() context.RouteReadOnly              { return nil }
func (mock *ContextMock) Do(context.Handlers)                                 {}
func (mock *ContextMock) AddHandler(...context.Handler)                       {}
func (mock *ContextMock) SetHandlers(context.Handlers)                        {}
func (mock *ContextMock) Handlers() context.Handlers                          { return nil }
func (mock *ContextMock) HandlerIndex(_ int) int                              { return 0 }
func (mock *ContextMock) Proceed(context.Handler) bool                        { return false }
func (mock *ContextMock) HandlerName() string                                 { return "" }
func (mock *ContextMock) HandlerFileLine() (string, int)                      { return "", 0 }
func (mock *ContextMock) RouteName() string                                   { return "" }
func (mock *ContextMock) Next()                                               {}
func (mock *ContextMock) NextOr(_ ...context.Handler) bool                    { return false }
func (mock *ContextMock) NextOrNotFound() bool                                { return false }
func (mock *ContextMock) NextHandler() context.Handler                        { return nil }
func (mock *ContextMock) Skip()                                               {}
func (mock *ContextMock) StopExecution()                                      {}
func (mock *ContextMock) IsStopped() bool                                     { return false }
func (mock *ContextMock) OnConnectionClose(_ func()) bool                     { return false }
func (mock *ContextMock) OnClose(_ func())                                    {}
func (mock *ContextMock) Params() *context.RequestParams                      { return nil }
func (mock *ContextMock) Values() *memstore.Store                             { return nil }
func (mock *ContextMock) Method() string                                      { return "" }
func (mock *ContextMock) Path() string                                        { return "" }
func (mock *ContextMock) RequestPath(_ bool) string                           { return "" }
func (mock *ContextMock) Host() string                                        { return "" }
func (mock *ContextMock) Subdomain() string                                   { return "" }
func (mock *ContextMock) FindClosest(_ int) []string                          { return nil }
func (mock *ContextMock) IsWWW() bool                                         { return false }
func (mock *ContextMock) FullRequestURI() string                              { return "" }
func (mock *ContextMock) RemoteAddr() string                                  { return "" }
func (mock *ContextMock) GetHeader(_ string) string                           { return "" }
func (mock *ContextMock) IsAjax() bool                                        { return false }
func (mock *ContextMock) IsMobile() bool                                      { return false }
func (mock *ContextMock) IsScript() bool                                      { return false }
func (mock *ContextMock) GetReferrer() context.Referrer                       { return context.Referrer{} }
func (mock *ContextMock) GetLocale() context.Locale                           { return nil }
func (mock *ContextMock) Tr(_ string, _ ...interface{}) string                { return "" }
func (mock *ContextMock) Header(_ string, _ string)                           {}
func (mock *ContextMock) ContentType(_ string)                                {}
func (mock *ContextMock) GetContentType() string                              { return "" }
func (mock *ContextMock) GetContentTypeRequested() string                     { return "" }
func (mock *ContextMock) GetContentLength() int64                             { return 0 }
func (mock *ContextMock) StatusCode(_ int)                                    {}
func (mock *ContextMock) GetStatusCode() int                                  { return 0 }
func (mock *ContextMock) AbsoluteURI(_ string) string                         { return "" }
func (mock *ContextMock) Redirect(_ string, _ ...int)                         {}
func (mock *ContextMock) URLParamExists(_ string) bool                        { return false }
func (mock *ContextMock) URLParamDefault(_ string, _ string) string           { return "" }
func (mock *ContextMock) URLParam(_ string) string                            { return "" }
func (mock *ContextMock) URLParamTrim(_ string) string                        { return "" }
func (mock *ContextMock) URLParamEscape(_ string) string                      { return "" }
func (mock *ContextMock) URLParamInt(_ string) (int, error)                   { return 0, nil }
func (mock *ContextMock) URLParamIntDefault(_ string, _ int) int              { return 0 }
func (mock *ContextMock) URLParamInt32Default(_ string, _ int32) int32        { return 0 }
func (mock *ContextMock) URLParamInt64(_ string) (int64, error)               { return 0, nil }
func (mock *ContextMock) URLParamInt64Default(_ string, _ int64) int64        { return 0 }
func (mock *ContextMock) URLParamFloat64(_ string) (float64, error)           { return 0, nil }
func (mock *ContextMock) URLParamFloat64Default(_ string, _ float64) float64  { return 0 }
func (mock *ContextMock) URLParamBool(_ string) (bool, error)                 { return false, nil }
func (mock *ContextMock) URLParams() map[string]string                        { return nil }
func (mock *ContextMock) FormValueDefault(_ string, _ string) string          { return "" }
func (mock *ContextMock) FormValue(_ string) string                           { return "" }
func (mock *ContextMock) FormValues() map[string][]string                     { return nil }
func (mock *ContextMock) PostValueDefault(_ string, _ string) string          { return "" }
func (mock *ContextMock) PostValue(_ string) string                           { return "" }
func (mock *ContextMock) PostValueTrim(_ string) string                       { return "" }
func (mock *ContextMock) PostValueInt(_ string) (int, error)                  { return 0, nil }
func (mock *ContextMock) PostValueIntDefault(_ string, _ int) int             { return 0 }
func (mock *ContextMock) PostValueInt64(_ string) (int64, error)              { return 0, nil }
func (mock *ContextMock) PostValueInt64Default(_ string, _ int64) int64       { return 0 }
func (mock *ContextMock) PostValueFloat64(_ string) (float64, error)          { return 0, nil }
func (mock *ContextMock) PostValueFloat64Default(_ string, _ float64) float64 { return 0 }
func (mock *ContextMock) PostValueBool(_ string) (bool, error)                { return false, nil }
func (mock *ContextMock) PostValues(_ string) []string                        { return nil }
func (mock *ContextMock) FormFile(_ string) (multipart.File, *multipart.FileHeader, error) {
	return nil, nil, nil
}
func (mock *ContextMock) UploadFormFiles(_ string, _ ...func(context.Context, *multipart.FileHeader)) (int64, error) {
	return 0, nil
}
func (mock *ContextMock) NotFound()                                                {}
func (mock *ContextMock) SetMaxRequestBodySize(_ int64)                            {}
func (mock *ContextMock) GetBody() ([]byte, error)                                 { return nil, nil }
func (mock *ContextMock) UnmarshalBody(_ interface{}, _ context.Unmarshaler) error { return nil }
func (mock *ContextMock) ReadJSON(_ interface{}) error                             { return nil }
func (mock *ContextMock) ReadXML(_ interface{}) error                              { return nil }
func (mock *ContextMock) ReadYAML(_ interface{}) error                             { return nil }
func (mock *ContextMock) ReadForm(_ interface{}) error                             { return nil }
func (mock *ContextMock) ReadQuery(_ interface{}) error                            { return nil }
func (mock *ContextMock) Write(_ []byte) (int, error)                              { return 0, nil }
func (mock *ContextMock) Writef(_ string, _ ...interface{}) (int, error)           { return 0, nil }
func (mock *ContextMock) WriteString(_ string) (int, error)                        { return 0, nil }
func (mock *ContextMock) SetLastModified(_ time.Time)                              {}
func (mock *ContextMock) CheckIfModifiedSince(_ time.Time) (bool, error)           { return false, nil }
func (mock *ContextMock) WriteNotModified()                                        {}
func (mock *ContextMock) WriteWithExpiration(_ []byte, _ time.Time) (int, error)   { return 0, nil }
func (mock *ContextMock) StreamWriter(_ func(w io.Writer) bool)                    {}
func (mock *ContextMock) ClientSupportsGzip() bool                                 { return false }
func (mock *ContextMock) WriteGzip(_ []byte) (int, error)                          { return 0, nil }
func (mock *ContextMock) TryWriteGzip(_ []byte) (int, error)                       { return 0, nil }
func (mock *ContextMock) GzipResponseWriter() *context.GzipResponseWriter          { return nil }
func (mock *ContextMock) Gzip(_ bool)                                              {}
func (mock *ContextMock) ViewLayout(_ string)                                      {}
func (mock *ContextMock) ViewData(_ string, _ interface{})                         {}
func (mock *ContextMock) GetViewData() map[string]interface{}                      { return nil }
func (mock *ContextMock) View(_ string, _ ...interface{}) error                    { return nil }
func (mock *ContextMock) Binary(_ []byte) (int, error)                             { return 0, nil }
func (mock *ContextMock) Text(_ string, _ ...interface{}) (int, error)             { return 0, nil }
func (mock *ContextMock) HTML(_ string, _ ...interface{}) (int, error)             { return 0, nil }
func (mock *ContextMock) JSON(_ interface{}, _ ...context.JSON) (int, error)       { return 0, nil }
func (mock *ContextMock) JSONP(_ interface{}, _ ...context.JSONP) (int, error)     { return 0, nil }
func (mock *ContextMock) XML(_ interface{}, _ ...context.XML) (int, error)         { return 0, nil }
func (mock *ContextMock) Problem(_ interface{}, _ ...context.ProblemOptions) (int, error) {
	return 0, nil
}
func (mock *ContextMock) Markdown(_ []byte, _ ...context.Markdown) (int, error) { return 0, nil }
func (mock *ContextMock) YAML(_ interface{}) (int, error)                       { return 0, nil }
func (mock *ContextMock) Negotiation() *context.NegotiationBuilder              { return nil }
func (mock *ContextMock) Negotiate(_ interface{}) (int, error)                  { return 0, nil }
func (mock *ContextMock) ServeContent(_ io.ReadSeeker, _ string, _ time.Time, _ bool) error {
	return nil
}
func (mock *ContextMock) ServeFile(_ string, _ bool) error                          { return nil }
func (mock *ContextMock) SendFile(_ string, _ string) error                         { return nil }
func (mock *ContextMock) SetCookie(_ *http.Cookie, _ ...context.CookieOption)       {}
func (mock *ContextMock) SetCookieKV(_ string, _ string, _ ...context.CookieOption) {}
func (mock *ContextMock) GetCookie(_ string, _ ...context.CookieOption) string      { return "" }
func (mock *ContextMock) RemoveCookie(_ string, _ ...context.CookieOption)          {}
func (mock *ContextMock) VisitAllCookies(_ func(_ string, _ string))                {}
func (mock *ContextMock) MaxAge() int64                                             { return 0 }
func (mock *ContextMock) Record()                                                   {}
func (mock *ContextMock) Recorder() *context.ResponseRecorder                       { return nil }
func (mock *ContextMock) IsRecording() (*context.ResponseRecorder, bool)            { return nil, false }
func (mock *ContextMock) BeginTransaction(_ func(t *context.Transaction))           {}
func (mock *ContextMock) SkipTransactions()                                         {}
func (mock *ContextMock) TransactionsSkipped() bool                                 { return false }
func (mock *ContextMock) Exec(_ string, _ string)                                   {}
func (mock *ContextMock) RouteExists(_ string, _ string) bool                       { return false }
func (mock *ContextMock) ReflectValue() []reflect.Value                             { return nil }
func (mock *ContextMock) Application() context.Application                          { return nil }
func (mock *ContextMock) String() string                                            { return "" }

type structWithEmptyInterface struct {
	EmptyInterface interface{}
}

func (_ structWithEmptyInterface) M1(interface{}) {
}

func TestIsEmptyInterface(t *testing.T) {

	theType := reflect.TypeOf(structWithEmptyInterface{}).Field(0).Type
	if !IsEmptyInterface(theType) {
		t.Fatalf(fmt.Sprintf("判断错误!实际类型为%s", theType.Name()))
	}
	theType = reflect.TypeOf(structWithEmptyInterface{}).Method(0).Type.In(1)
	if !IsEmptyInterface(theType) {
		t.Fatalf(fmt.Sprintf("判断错误!实际类型为%s", theType.Name()))
	}
}
