package grmi

import (
	"LEIT_PM/container"
	"errors"
	"fmt"
	"github.com/kataras/iris/v12"
	"github.com/kataras/iris/v12/core/router"
	"reflect"
	"strconv"
	"strings"
	"time"
)

type GoType int

const (
	URLDateTimeFormat = time.RFC3339Nano
	URLDateFormat     = "2006-01-02Z07:00"
	DateOutFormat     = "2006-01-02"
	DatetimeOutFormat = "2006-01-02 15:04:05"
)

const (
	TypeInt GoType = iota + 1
	TypeInt8
	TypeInt16
	TypeInt32
	TypeInt64
	TypeUint
	TypeUint8
	TypeUint16
	TypeUint32
	TypeUint64
	TypeFloat32
	TypeFloat64
	TypeDate
	TypeDateTime
	TypeBool
	TypeString
)

type Source string

type ParameterParser = func(iris.Context) (interface{}, error)

type ParameterParserBuilder = func(string) ParameterParser

const (
	User    Source = ""
	Session Source = "session"
)

// 分页结果 类型常量
var PagingResultType = reflect.TypeOf(&PagingResult{})

// float64 类型常量
var Float64Type = reflect.TypeOf(float64(0))

// DateTime 类型常量
var DateTimeType = reflect.TypeOf(DateTime{})

// Date 类型常量
var DateType = reflect.TypeOf(Date{})

var routeMapping = make(map[string]func(router.Party), 10)

var modelMapping = make(map[string]map[string]*Model, 10)

func RepeatStrings(content string, count int) []string {
	result := make([]string, count)
	for index := range result {
		result[index] = content
	}
	return result
}

func Equal(predicate Predicate) (string, error) {
	if predicate.Values == nil || len(predicate.Values) != 1 {
		return "", errors.New("相等性比较只支持一个参数")
	}
	return fmt.Sprintf(" and %s = ?", predicate.ColumnName), nil
}

func NotEqual(predicate Predicate) (string, error) {
	if predicate.Values == nil || len(predicate.Values) != 1 {
		return "", errors.New("相等性比较只支持一个参数")
	}
	return fmt.Sprintf(" and %s <> ?", predicate.ColumnName), nil
}

func Like(predicate Predicate) (string, error) {
	if predicate.Values == nil || len(predicate.Values) != 1 {
		return "", errors.New("相似性比较只支持一个参数")
	}
	return fmt.Sprintf(" and %s like concat('%%', ?, '%%')", predicate.ColumnName), nil
}

func NotLike(predicate Predicate) (string, error) {
	if predicate.Values == nil || len(predicate.Values) != 1 {
		return "", errors.New("相似性比较只支持一个参数")
	}
	return fmt.Sprintf(" and %s not like concat('%%', ?, '%%')", predicate.ColumnName), nil
}

func IsNull(predicate Predicate) (string, error) {
	if predicate.Values != nil && len(predicate.Values) > 0 {
		return "", errors.New("为空比较比较不支持参数")
	}
	return fmt.Sprintf(" and %s is null", predicate.ColumnName), nil
}

func IsNotNull(predicate Predicate) (string, error) {
	if predicate.Values != nil && len(predicate.Values) > 0 {
		return "", errors.New("为空比较比较不支持参数")
	}
	return fmt.Sprintf(" and %s is not null", predicate.ColumnName), nil
}

func Include(predicate Predicate) (string, error) {
	if predicate.Values != nil && len(predicate.Values) > 0 {
		return fmt.Sprintf(" and %s in (%s)", predicate.ColumnName, strings.Join(RepeatStrings("?", len(predicate.Values)), ",")), nil
	} else {
		return " and 1 <> 1", nil
	}
}

func Exclude(predicate Predicate) (string, error) {
	if predicate.Values != nil && len(predicate.Values) > 0 {
		return fmt.Sprintf(" and %s not in (%s)", predicate.ColumnName, strings.Join(RepeatStrings("?", len(predicate.Values)), ",")), nil
	} else {
		return " and 1 = 1", nil
	}
}

func Approximate(predicate Predicate) (string, error) {
	if predicate.Values == nil || len(predicate.Values) != 1 {
		return "", errors.New("约等于比较比较只支持一个参数")
	}
	return fmt.Sprintf(" and datediff(second, %s, ?) = 0", predicate.ColumnName), nil
}

func NotApproximate(predicate Predicate) (string, error) {
	if predicate.Values == nil || len(predicate.Values) != 1 {
		return "", errors.New("约等于比较比较只支持一个参数")
	}
	return fmt.Sprintf(" and datediff(second, %s, ?) <> 0", predicate.ColumnName), nil
}

func GreaterThan(predicate Predicate) (string, error) {
	if predicate.Values == nil || len(predicate.Values) != 1 {
		return "", errors.New("大于比较比较只支持一个参数")
	}
	return fmt.Sprintf(" and %s > ?", predicate.ColumnName), nil
}

func GreaterOrEqual(predicate Predicate) (string, error) {
	if predicate.Values == nil || len(predicate.Values) != 1 {
		return "", errors.New("大于等于比较比较只支持一个参数")
	}
	return fmt.Sprintf(" and %s >= ?", predicate.ColumnName), nil
}

func LessThan(predicate Predicate) (string, error) {
	if predicate.Values == nil || len(predicate.Values) != 1 {
		return "", errors.New("小于比较比较只支持一个参数")
	}
	return fmt.Sprintf(" and %s < ?", predicate.ColumnName), nil
}

func LessOrEqual(predicate Predicate) (string, error) {
	if predicate.Values == nil || len(predicate.Values) != 1 {
		return "", errors.New("小于等于比较比较只支持一个参数")
	}
	return fmt.Sprintf(" and %s <= ?", predicate.ColumnName), nil
}

func intParser(value string) (interface{}, error) {
	predicateValue, err := strconv.ParseInt(value, 10, 64)
	if err != nil {
		return nil, err
	}
	return predicateValue, nil
}

func uint64Parser(value string) (interface{}, error) {
	predicateValue, err := strconv.ParseUint(value, 10, 64)
	if err != nil {
		return nil, err
	}
	return predicateValue, nil
}

func floatParser(value string) (interface{}, error) {
	predicateValue, err := strconv.ParseFloat(value, 64)
	if err != nil {
		return nil, err
	}
	return predicateValue, nil
}

func dateTimeParser(value string) (interface{}, error) {
	predicateValue, err := time.Parse(URLDateTimeFormat, value)
	if err != nil {
		return nil, err
	}
	return predicateValue, nil
}

func dateParser(value string) (interface{}, error) {
	predicateValue, err := time.Parse(URLDateFormat, value)
	if err != nil {
		return nil, err
	}
	return predicateValue, nil
}

func boolParser(value string) (interface{}, error) {
	predicateValue, err := strconv.ParseBool(value)
	if err != nil {
		return nil, err
	}
	return predicateValue, nil
}

func stringParser(value string) (interface{}, error) {
	return strings.TrimSpace(value), nil
}

var valueParsers = map[GoType]func(string) (interface{}, error){
	TypeInt:      intParser,
	TypeInt8:     intParser,
	TypeInt16:    intParser,
	TypeInt32:    intParser,
	TypeInt64:    intParser,
	TypeUint:     intParser,
	TypeUint8:    intParser,
	TypeUint16:   intParser,
	TypeUint32:   intParser,
	TypeUint64:   uint64Parser,
	TypeFloat32:  floatParser,
	TypeFloat64:  floatParser,
	TypeDateTime: dateTimeParser,
	TypeDate:     dateParser,
	TypeBool:     boolParser,
	TypeString:   stringParser,
}

var basicTypes = map[GoType]reflect.Type{
	TypeInt:      container.IntType,
	TypeInt8:     container.Int8Type,
	TypeInt16:    container.Int16Type,
	TypeInt32:    container.Int32Type,
	TypeInt64:    container.Int64ype,
	TypeUint:     container.UintType,
	TypeUint8:    container.Uint8Type,
	TypeUint16:   container.Uint16Type,
	TypeUint32:   container.Uint32Type,
	TypeUint64:   container.Uint64Type,
	TypeFloat32:  container.Float32Type,
	TypeFloat64:  container.Float64Type,
	TypeDateTime: DateTimeType,
	TypeDate:     DateType,
	TypeBool:     container.BoolType,
	TypeString:   container.StringType,
}

var parameterParserBuilders = map[GoType]ParameterParserBuilder{
	TypeInt: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetInt(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeInt8: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetInt8(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeInt16: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetInt16(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeInt32: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetInt32(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeInt64: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetInt64(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeUint: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetUint(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeUint8: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetUint8(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeUint16: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetUint16(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeUint32: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetUint32(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeUint64: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetUint64(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeFloat32: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			originValue, err := context.Params().GetFloat64(parameterName)
			if err != nil {
				return nil, err
			}
			return float32(originValue), nil
		}
	},
	TypeFloat64: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetFloat64(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeDateTime: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			originValue, err := time.Parse(URLDateTimeFormat, context.Params().GetString(parameterName))
			if err != nil {
				return nil, err
			}
			return DateTime(originValue), nil
		}
	},
	TypeDate: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			originValue, err := time.Parse(URLDateFormat, context.Params().GetString(parameterName))
			if err != nil {
				return nil, err
			}
			return Date(originValue), nil
		}
	},
	TypeBool: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			value, err := context.Params().GetBool(parameterName)
			if err != nil {
				return nil, err
			}
			return value, nil
		}
	},
	TypeString: func(parameterName string) ParameterParser {
		return func(context iris.Context) (interface{}, error) {
			return context.Params().GetString(parameterName), nil
		}
	},
}

func Models() map[string]map[string]*Model {
	return modelMapping
}

func RegisterEntities(moudleName string, models ...*Model) {
	moudleMapping, ok := modelMapping[moudleName]
	if !ok {
		moudleMapping = make(map[string]*Model, 50)
		modelMapping[moudleName] = moudleMapping
	}
	for _, entity := range models {

		moudleMapping[strings.ToLower(entity.ModelType.Name())] = entity
	}
}
